# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['tokamak', 'tokamak.radix_tree', 'tokamak.web']

package_data = \
{'': ['*']}

extras_require = \
{'examples': ['hypercorn>=0.13.2,<0.14.0', 'starlette>=0.20.4,<0.21.0'],
 'full': ['hypercorn>=0.13.2,<0.14.0', 'trio>=0.20.0,<0.21.0'],
 'web': ['trio>=0.20.0,<0.21.0']}

setup_kwargs = {
    'name': 'tokamak',
    'version': '0.5.3',
    'description': 'HTTP Router based on radix trees',
    'long_description': '# Tokamak\n\n[![PyPI version fury.io](https://badge.fury.io/py/tokamak.svg)](https://pypi.python.org/pypi/tokamak/)\n[![Documentation Status](https://readthedocs.org/projects/tokamak/badge/?version=latest)](https://tokamak.readthedocs.io/en/latest/?badge=latest)\n\nTokamak is a router based on Radix trees intended for ASGI Python applications.\n\n[[Read the documentation](https://tokamak.readthedocs.io/en/latest/)]\n\n## Primary Project Goals\n\nThere are many HTTP routers based on radix trees available in other language communities, including Go, Javascript, Rust, and others. In Python, however, most open-source Python web frameworks instead utilize lists to store and look up HTTP routes.\n\nThe primary goal for this project is to provide a radix-tree-based router for Python web frameworks (or any custom ASGI or WSGI implementation).\n\n**This library is _experimental_. Use at your own risk.**\n\n### Other Goals\n\nAs a secondary goal, a minimal and highly experimental web framework is included an optional install in this library.\n\nThe reasons for building and including this web framework are as follows:\n\n- It provides a convenient way to test the [`AsgiRouter`](/routing) class, and\n- It allows the authors of this library to explore experimental ASGI-framework features, including request-cancellation, background task time-limits, and background task cancellation.\n\nDevelopers should consider more fully-featured web frameworks _before_ this one and there are many to choose from:\n\n- [Django](https://www.djangoproject.com/)\n- [Flask](https://flask.palletsprojects.com/en/2.2.x/)\n- [Starlette](https://www.starlette.io/)\n- [Tornado](https://www.tornadoweb.org/en/stable/)\n- [Quart](https://pgjones.gitlab.io/quart/)\n\n## Installation\n\nYou can install `tokamak` with:\n\n```sh\npip install tokamak\n```\n\nBy default tokamak has no dependencies. If you would like to try out the experimental web framework, you can install with optional extras `web`, which will include `trio`:\n\n```sh\n$ pip install "tokamak[web]"\n...\n```\n\n## Usage\n\nThis library provides a radix tree implementation and a basic `AsgiRouter` router implementation for low-level ASGI applications. You can use the `AsgiRouter` class as follows.\n\nFirst, we start with some with some imports and some fallback handlers:\n\n```python\nfrom hypercorn.config import Config\nfrom hypercorn.trio import serve\nimport trio\n\nfrom tokamak import AsgiRouter, Route\nfrom tokamak.router import MethodNotAllowed, UnknownEndpoint\n\n\n# # Fallback Handlers # #\nasync def method_not_allowed(scope, receive, send):\n    await send(\n        {\n            "type": "http.response.start",\n            "status": 405,\n            "headers": [(b"Content-Type", b"text/html; charset=UTF-8")],\n        }\n    )\n    await send(\n        {\n            "type": "http.response.body",\n            "body": b"<html><body><h1>405 Method not allowed</h1></body></html>",\n        }\n    )\n\n\nasync def unknown_handler(scope, receive, send):\n    await send(\n        {\n            "type": "http.response.start",\n            "status": 404,\n            "headers": [(b"Content-Type", b"text/html; charset=UTF-8")],\n        }\n    )\n    await send(\n        {\n            "type": "http.response.body",\n            "body": b"<html><body><h1>404 Not Found!</h1></body></html>",\n        }\n    )\n```\n\nNext we\'ll build two different application endpoint handlers. These do roughly the same thing, so this is purely for demonstration purposes:\n\n```python\nasync def index(path_context, scope, receive, send):\n    message = await receive()\n    if message["type"] == "http.request":\n        body = message.get("body", b"")\n        # here\'s our response:\n        await send({"type": "http.response.start", "status": 200})\n        await send({"type": "http.response.body", "body": body if body else b"OK"})\n    elif message["type"] == "http.disconnect":\n        print("Disconnected! ")\n\n\nasync def other_handler(path_context, scope, receive, send):\n    context = bytes(json.dumps(path_context), encoding="utf-8")\n    message = await receive()\n    if message["type"] == "http.request":\n        body = message.get("body", b"")\n        await send({"type": "http.response.start", "status": 200})\n        await send({"type": "http.response.body", "body": body if body else context})\n    elif message["type"] == "http.disconnect":\n        print("Disconnected! ")\n```\n\nFinally, we can build an `AsgiRouter` and a working ASGI app, like this:\n\n```python\n# `AsgiRouter` and `Route` class provided by this library\nROUTER = AsgiRouter(\n    routes=[\n        Route("/", handler=index, methods=["GET"]),\n        # Routes will match on regexes and bind to variables\n        # given on the left side of the colon\n        Route(\n            "/other_handler/{name:[a-z1-9]+}", handler=other_handler, methods=["POST"],\n        ),\n    ]\n)\n\n# This is a basic implementation of the ASGI spec\n# See: https://asgi.readthedocs.io/en/latest/specs/main.html\nasync def asgi_app(scope, receive, send):\n    path = scope.get("path", "")\n    try:\n        # Routers provider a `get_route` method\n        # If no route is matched, they throw `UnknownEndpoint`\n        # If a route is matched, we\'ll get path context and a handler\n        handler, context = ROUTER.get_route(path)\n    except UnknownEndpoint:\n        await unknown_handler(scope, receive, send)\n        return None\n\n    try:\n        # If a matched router doesn\'t handle this method\n        # it will throw `MethodNotAllowed`\n        await handler(context, scope, receive, send, method=scope.get("method"))\n    except MethodNotAllowed:\n      await method_not_allowed(scope, receive, send)\n      return None\n\n\nasync def app_with_lifespan(scope, receive, send):\n    if scope["type"] == "lifespan":\n        while True:\n            message = await receive()\n            if message["type"] == "lifespan.startup":\n                await send({"type": "lifespan.startup.complete"})\n            elif message["type"] == "lifespan.shutdown":\n                await send({"type": "lifespan.shutdown.complete"})\n                return\n    if scope["type"] == "http":\n        return await asgi_app(scope, receive, send)\n```\n\nFinally, to run our ASGI app, we\'ll add the following:\n\n```python\n\nif __name__ == "__main__":\n    config = Config()\n    config.bind = ["localhost:8000"]\n    trio.run(partial(serve, app_with_lifespan, config))\n```\n\nThis example relies on the following dependencies:\n\n- hypercorn\n- trio\n\nIf we have these dependencies in our Python environment, we can this simple script:\n\n```sh\n$ poetry run python examples/asgi_minimal.py\n[2022-03-20 16:59:58 -0700] [91988] [INFO] Running on http://127.0.0.1:8000 (CTRL + C to quit)\n```\n\nIn a separate terminal, we can try it out like so:\n\n```sh\n❯ curl http://localhost:8000/\nOK\n\n# No capital letters matched\n❯ curl -XPOST http://localhost:8000/other_handler/bla1AA\n<html><body><h1>404 Not Found!</h1></body></html>\n\n# GET not POST -> 405\n❯ curl http://localhost:8000/other_handler/bla\n<html><body><h1>405 Method not allowed</h1></body></html>\n\n# Success\n❯ curl -XPOST http://localhost:8000/other_handler/bla1\n{"name": "bla1"}\n```\n\n**Note**: that our regex path _does not_ match capital letters, so that request 404s.\n\n## Examples\n\nRunnable examples are provided in the [`examples` directory](https://github.com/erewok/tokamak/examples). For instance, you can run the experimental `tokamak` application with `trio` and `hypercorn` like so:\n\n```sh\n$ poetry install -E "full"\nInstalling dependencies from lock file\n\nPackage operations: 8 installs, 0 updates, 0 removals\n\n  • Installing h11 (0.13.0)\n  • Installing hpack (4.0.0)\n  • Installing hyperframe (6.0.1)\n  • Installing h2 (4.1.0)\n  • Installing priority (2.0.0)\n  • Installing toml (0.10.2)\n  • Installing wsproto (1.1.0)\n  • Installing hypercorn (0.13.2)\n\nInstalling the current project: tokamak (0.2.1)\n❯ poetry run python examples/tokamak_app.py\n========·°·°~> Starting tokamak °°···°°🚀···°°\n[2022-03-20 11:05:01 -0700] [63023] [INFO] Running on http://127.0.0.1:8000 (CTRL + C to quit)\n```\n\nIn a separate terminal, you can make various requests, such as the following:\n\n```sh\n❯ curl http://localhost:8000\n{"received": {}}\n\n❯ curl http://localhost:8000/info/erik -d \'{"some_data": "something"}\'\n{"received": {"some_data": "something"}}\n```\n\nBack in the first terminal, where you launched the example `tokamak` application, you should see the following:\n\n```sh\n❯ poetry run python examples/tokamak_app.py\n========·°·°~> Starting tokamak °°···°°🚀···°°\n[2022-03-20 11:05:01 -0700] [63023] [INFO] Running on http://127.0.0.1:8000 (CTRL + C to quit)\n{} {\'type\': \'http\', \'http_version\': \'1.1\', \'asgi\': {\'spec_version\': \'2.1\', \'version\': \'3.0\'}, \'method\': \'GET\', \'scheme\': \'http\', \'path\': \'/\', \'raw_path\': b\'/\', \'query_string\': b\'\', \'root_path\': \'\', \'headers\': <Headers([(b\'host\', b\'localhost:8000\'), (b\'user-agent\', b\'curl/7.81.0\'), (b\'accept\', b\'*/*\')])>, \'client\': (\'127.0.0.1\', 55379), \'server\': (\'127.0.0.1\', 8000), \'extensions\': {}} <Headers([(b\'host\', b\'localhost:8000\'), (b\'user-agent\', b\'curl/7.81.0\'), (b\'accept\', b\'*/*\')])> b\'\' 1.1 GET\nSleeping 1s for total seconds: 0\nSleeping 1s for total seconds: 1\nSleeping 1s for total seconds: 2\nSleeping 1s for total seconds: 3\nSleeping 1s for total seconds: 4\n{\'user\': \'erik\'} {\'type\': \'http\', \'http_version\': \'1.1\', \'asgi\': {\'spec_version\': \'2.1\', \'version\': \'3.0\'}, \'method\': \'POST\', \'scheme\': \'http\', \'path\': \'/info/erik\', \'raw_path\': b\'/info/erik\', \'query_string\': b\'\', \'root_path\': \'\', \'headers\': <Headers([(b\'host\', b\'localhost:8000\'), (b\'user-agent\', b\'curl/7.81.0\'), (b\'accept\', b\'*/*\'), (b\'content-length\', b\'26\'), (b\'content-type\', b\'application/x-www-form-urlencoded\')])>, \'client\': (\'127.0.0.1\', 55386), \'server\': (\'127.0.0.1\', 8000), \'extensions\': {}} <Headers([(b\'host\', b\'localhost:8000\'), (b\'user-agent\', b\'curl/7.81.0\'), (b\'accept\', b\'*/*\'), (b\'content-length\', b\'26\'), (b\'content-type\', b\'application/x-www-form-urlencoded\')])> b\'\' 1.1 POST\nSleeping 1s for total seconds: 0\nSleeping 1s for total seconds: 5\nSleeping 1s for total seconds: 1\nSleeping 1s for total seconds: 6\nSleeping 1s for total seconds: 2\nSleeping 1s for total seconds: 7\nSleeping 1s for total seconds: 3\nSleeping 1s for total seconds: 8\n```\n',
    'author': 'Erik Aker',
    'author_email': 'eraker@gmail.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'packages': packages,
    'package_data': package_data,
    'extras_require': extras_require,
    'python_requires': '>=3.9,<4.0',
}


setup(**setup_kwargs)
